//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#include "Oinject.h"

ClassImp(Oinject)

////////////////////////////////////////////////////////////////////////////////////
Oinject::Oinject(const double aDuration){ 
////////////////////////////////////////////////////////////////////////////////////

  duration = aDuration;
  
  taumin  = 0;
  taumax  = 0;
  phimin  = 32;
  phimax  = 2048;
  Qmin    = 4;
  Qmax    = 100;
  ampmin  = 1e-21;
  ampmax  = 1e-21;

  // random generator
  randgen = new TRandom3();
  randgen->SetSeed(0); // random seed
  MakeWaveform();
  
}

////////////////////////////////////////////////////////////////////////////////////
Oinject::~Oinject(void){
////////////////////////////////////////////////////////////////////////////////////
  delete randgen;
}

////////////////////////////////////////////////////////////////////////////////////
void Oinject::MakeWaveform(void){
////////////////////////////////////////////////////////////////////////////////////

  // waveform parameters
  GenerateParameters();

  // derived parameters
  Wg       = sqrt(sqrt(2.0/TMath::Pi())*Q/phi);
  sigma_t = Q/phi/TMath::Pi()/sqrt(8.0);
  sigma_f = sqrt(2.0)*phi/Q;

  return;
}

////////////////////////////////////////////////////////////////////////////////////
double Oinject::GetTrueSNR(Spectrum *aSpec1, Spectrum *aSpec2){
////////////////////////////////////////////////////////////////////////////////////

  double freq, win, sum=0;
  double dfreq=aSpec1->GetSpectrumResolution();

  // only positive frequencies. No negative frequency contribution w=0.
  for(int i=1; i<aSpec1->GetSpectrumSize(); i++){
    freq=aSpec1->GetSpectrumFrequency(i);
    win = Wg * exp(-(freq-phi)*(freq-phi)/2.0/sigma_f/sigma_f);
    sum += win*win/2.0 / sqrt(aSpec1->GetPower(freq)*aSpec2->GetPower(freq)/4.0) * dfreq;
  }

  return amp*sum;
}

////////////////////////////////////////////////////////////////////////////////////
void Oinject::GenerateParameters(void){
////////////////////////////////////////////////////////////////////////////////////

  tau   = randgen->Uniform(taumin,taumax);
  phase = randgen->Uniform(0,2*TMath::Pi());
  amp   = ampmin*pow(ampmax/ampmin,randgen->Uniform());
  phi   = randgen->Uniform(phimin, phimax);
  Q     = randgen->Uniform(Qmin, Qmax);


}
